from __future__ import annotations

import typing
from enum import Enum
from functools import cached_property

from pydantic import BaseModel, Field, field_serializer, conlist, HttpUrl


class BilibiliResponse(BaseModel):
    code: int
    data: typing.Any
    message: str | None = None

    def raise_for_status(self) -> None:
        if self.code != 0:
            raise ValueError(
                f"Response failed with code: {self.code}. Message: {self.message}"
            )


class BilibiliVideoResponse(BilibiliResponse):
    """Base response model of bilibili api with video data return"""

    data: BilibiliVideoResponseData


class BilibiliPlayResponse(BilibiliResponse):
    """Base response model of bilibili api with play data return"""

    data: BilibiliPlayResponseData


class BilibiliVideoResponseData(BaseModel):
    """Base data model of bilibili api"""

    list_data: typing.Sequence[VideoItem] = Field([], validation_alias="list")


class BilibiliPlayResponseData(BaseModel):
    list_data: typing.Sequence[PlayItem] = Field([], validation_alias="list")


class RankDramaResponse(BilibiliPlayResponse):
    data: RankDramaData = Field(..., validation_alias="result")


class PopularResponse(BilibiliVideoResponse):
    """Response model of bilibili popular api"""

    data: PopularData


class WeeklyResponse(BilibiliVideoResponse):
    """Response model of bilibili weekly api"""

    data: WeeklyData


class PreciousResponse(BilibiliVideoResponse):
    data: PreciousData


class AuthorVideoResponse(BilibiliVideoResponse):
    data: AuthorVideoData


class PopularData(BilibiliVideoResponseData):
    """Data model of bilibili popular api"""

    no_more: bool
    list_data: list[PopularVideoItem] = Field([], validation_alias="list")


class WeeklyData(BilibiliVideoResponseData):
    """Data model of bilibili weekly api"""

    config: dict[str, typing.Any]
    reminder: str
    list_data: list[WeeklyVideoItem] = Field([], validation_alias="list")


class PreciousData(BilibiliVideoResponseData):
    list_data: list[PreciousVideoItem] = Field([], validation_alias="list")
    explain: str
    media_id: int
    title: str


class RankDramaData(BilibiliPlayResponseData):
    list_data: list[PlayItem] = Field([], validation_alias="list")
    note: str


class _AuthorVideoDataList(BaseModel):
    items: list[AuthorVideoItem] = Field([], validation_alias="vlist")


class AuthorVideoData(BilibiliVideoResponseData):
    list_data: _AuthorVideoDataList = Field(
        {
            "vlist": [],
        },
        validation_alias="list",
    )  # type: ignore


class VideoOwner(BaseModel):
    mid: int
    name: str
    face: str


class VideoStat(BaseModel):
    aid: int
    coin: int
    danmaku: int
    dislike: int
    favorite: int
    his_rank: int
    like: int
    now_rank: int
    reply: int
    share: int
    view: int


class PlayStat(BaseModel):
    danmaku: int
    follow: int
    series_follow: int
    view: int


class VideoItem(BaseModel):
    aid: int
    bvid: str
    cid: int
    desc: str
    owner: VideoOwner
    pubdate: int
    short_link_v2: str
    stat: VideoStat
    tid: int
    title: str
    tname: str

    @field_serializer("owner", "stat")
    def to_string(self, value: BaseModel) -> str:
        return value.model_dump_json() if isinstance(value, BaseModel) else value


class PlayItem(BaseModel):
    rank: int
    rating: str
    stat: PlayStat
    title: str
    url: str

    @field_serializer("stat")
    def to_string(self, value: BaseModel) -> str:
        return value.model_dump_json() if isinstance(value, BaseModel) else value


class PopularVideoItem(VideoItem):
    pass


class WeeklyVideoItem(VideoItem):
    pass


class PreciousVideoItem(VideoItem):
    achievement: str


class AuthorVideoItem(BaseModel):
    title: str
    aid: int
    bvid: str
    mid: int
    comment: int
    description: str | None = None
    is_pay: int
    length: str


class VideoSource(Enum):
    POPULAR = 0
    WEEKLY = 1
    HISTORY = 2
    RANK = 3
    MUSIC = 4
    DRAMA = 5
    SEARCH = 6


class PlayInfoResponse(BaseModel):
    code: int
    message: str | None = None
    data: PlayInfoData


class PlayInfoData(BaseModel):
    accept_quality: list[int]
    accept_description: list[str]
    dash: PlayInfoDash

    @cached_property
    def quality_map(self) -> dict[int, str]:
        return dict(zip(self.accept_quality, self.accept_description))


class _PlayMediaInfo(BaseModel):
    base_url: str
    backup_url: list[str]


class PlayVideo(_PlayMediaInfo):
    quality: int = Field(..., validation_alias="id")
    codecs: str


class PlayAudio(_PlayMediaInfo):
    audio_id: int = Field(..., validation_alias="id")


class PlayInfoDash(BaseModel):
    video: conlist(PlayVideo, min_length=1)  # type: ignore
    audio: conlist(PlayAudio, min_length=1)  # type: ignore


class WbiInfo(BaseModel):
    img_url: HttpUrl
    sub_url: HttpUrl

    @cached_property
    def img_key(self) -> str:
        return self.img_url.path.split("/")[-1].split(".")[0]  # type: ignore

    @cached_property
    def sub_key(self) -> str:
        return self.sub_url.path.split("/")[-1].split(".")[0]  # type: ignore



from fake_useragent import UserAgent  # type: ignore

from pathlib import Path

ua = UserAgent()


def user_agent_headers() -> dict[str, str]:
    return {
        "User-Agent": ua.random,
    }


def rm_tree(pth: Path):
    pth = Path(pth)
    if not pth.exists():
        return
    for child in pth.glob("*"):
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)
    pth.rmdir()



import logging
import logging.config

from spiders_for_all.conf import settings

LEVEL = settings.LOG_LEVEL

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "rich": {
            "format": "%(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
        "file": {
            "format": "%(asctime)s %(levelname)s %(name)s %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
    },
    "handlers": {
        "default": {
            "level": LEVEL,
            "class": "rich.logging.RichHandler",
            "formatter": "rich",
        },
        "bilibili-file": {
            "level": LEVEL,
            "class": "logging.handlers.RotatingFileHandler",
            "filename": settings.WORKDIR / "logs" / "bilibili.log",
            "maxBytes": 1024 * 1024 * 10,
            "backupCount": 5,
            "encoding": "utf-8",
            "formatter": "file",
        },
    },
    "loggers": {
        "bilibili": {
            "handlers": ["default", "bilibili-file"],
            "level": LEVEL,
            "propagate": True,
        },
        "default": {
            "handlers": ["default"],
            "level": LEVEL,
        },
    },
}

logging.config.dictConfig(LOGGING_CONFIG)


def get_logger(name: str):
    return logging.getLogger(name)


default_logger = get_logger("default")



from pathlib import Path

import click
import sqlalchemy as sa
from rich import print

from spiders_for_all.bilibili import spiders, download
from spiders_for_all.bilibili import analysis
from spiders_for_all.bilibili import db
from spiders_for_all.core.base import SPIDERS

_ = spiders  # to call init_subclass


@click.group()
def cli():
    pass


@click.option("--name", "-n", help="Spider name", required=True)
@click.option(
    "--params",
    "-p",
    help="Spider params",
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
)
@click.option(
    "--save-dir",
    "-s",
    help="If set `--save-dir`, all the videos will be downloaded when the spider finished",
    type=Path,
    required=False,
)
@click.option("--sess-data", "-ss", help="`SESSDATA`", required=False)
@click.option(
    "--download-only",
    "-d",
    is_flag=True,
)
@cli.command()
def run_spider(
    name: str,
    params: tuple[tuple[str, str]],
    save_dir: Path | None = None,
    sess_data: str | None = None,
    download_only: bool = False,
):
    """Run a spider by name"""
    if name not in SPIDERS:
        raise ValueError(f"Spider {name} not found")  # pragma: no cover
    spider = SPIDERS[name](**{k: v for k, v in params})

    if not download_only:
        print(f"[bold light_green]Running spider: {spider.string()}")
        spider.run()

    if download_only and save_dir is None:
        click.echo("You must set `--save-path` to download the videos")
        exit(1)

    if save_dir is not None:
        with db.Session() as s:
            db_model = spider.database_model
            select_bvid_stmt = sa.select(getattr(db_model, "bvid"))
            bvid_list = [row.bvid for row in s.execute(select_bvid_stmt)]

            print(f"[bold yellow]{len(bvid_list)} videos found to be downloaded...")

            multiple_downloader = download.MultiThreadDownloader(
                bvid_list=bvid_list,
                save_dir=save_dir,
                sess_data=sess_data,
                from_cli=True,
                max_workers=4,
            )

            multiple_downloader.download()


@cli.command()
def list_spiders():
    """List all available spiders"""
    print("Available spiders:")
    reversed_map = {v: k for k, v in SPIDERS.items()}
    for spider in reversed_map:
        print(f"  - {spider.string()})")


@cli.command()
@click.option("--name", "-n", help="Spider name", required=True)
@click.option("--top-n", "-t", help="Top N", required=True, type=int, default=10)
def data_analysis(name: str, top_n: int):
    """List the data as table for a spider"""
    if name not in SPIDERS:
        raise ValueError(f"Spider {name} not found")  # pragma: no cover
    spider = SPIDERS[name]
    print(f"Running analysis: {spider.string()}")
    analysis.Analysis(spider.database_model, top_n).show()  # type: ignore


@cli.command()
@click.option("--bvid", "-b", help="Bvid of the video", required=True, type=str)
@click.option(
    "--save-dir", "-s", help="Save directory of the video", required=True, type=Path
)
@click.option(
    "--filename",
    "-f",
    help="Filename of the video. By default, will use the bvid as the output filename.",
    required=False,
    type=str,
)
@click.option(
    "--remove-temp-dir",
    "-r",
    help="Whether to remove the temp dir after download",
    required=False,
    type=bool,
    default=True,
)
@click.option(
    "--sess-data",
    "-d",
    help="Pass to requests as cookies to download high quality video. "
    "You should get this from:"
    "your browser -> Devtools -> Choose any request -> Cookies -> SESSDATA.",
    required=False,
    type=str,
)
@click.option(
    "--quality",
    "-q",
    help="Quality of the video to download. Defaults to HIGHEST_QUALITY.",
    required=False,
    type=int,
    default=download.HIGHEST_QUALITY,
)
@click.option(
    "--codecs",
    "-c",
    help="Regex to filter video codecs. Defaults to None.",
    required=False,
    type=str,
)
@click.option(
    "--ffmpeg-params",
    "-fp",
    help="Additional ffmpeg parameters. Defaults to None.",
    required=False,
    type=str,
    multiple=True,
)
def download_video(
    bvid: str,
    save_dir: str | Path,
    filename: str | None = None,
    remove_temp_dir: bool = True,
    sess_data: str | None = None,
    quality: int = download.HIGHEST_QUALITY,
    codecs: str | None = None,
    ffmpeg_params: list[str] | None = None,
):
    """Download a video by bvid"""
    downloader = download.Downloader(
        bvid,
        save_dir,
        filename,
        remove_temp_dir,
        sess_data,
        quality,
        codecs,
        ffmpeg_params,
    )

    with downloader:
        downloader.download()


@cli.command()
@click.option(
    "--bvids",
    "-b",
    help="bvid list provided to download the videos",
    required=True,
)
@click.option(
    "--save-dir",
    "-s",
    help="Path to save the downloaded videos",
    type=Path,
    required=True,
)
@click.option("--sess-data", "-ss", help="`SESSDATA`", required=False)
@click.option("--max-workers", "-w", help="max workers", type=int, default=4)
def download_videos(
    bvids: str | Path,
    save_dir: Path,
    sess_data: str | None = None,
    max_workers: int = 4,
):
    """Download multiple videos once

    You can do this by providing the list of bvid with `,`, `\n`, `\t`, or space seperated:

    >> python -m spiders_for_all.bilibili download-videos -b BVID1 BVID2 BVID3

    Or by providing a file contains bvid data like:

    \b
    bvid.txt:

    \b
    BVID1
    BVID2
    ...

    >> python -m spiders_for_all.bilibili download-videos -b bvid.txt
    """

    path_bvid = Path(bvids)

    if path_bvid.exists():
        bvids = path_bvid

    multiple_downloader = download.MultiThreadDownloader(
        bvid_list=bvids, save_dir=save_dir, sess_data=sess_data, max_workers=max_workers
    )

    multiple_downloader.download()


@cli.command()
@click.option("--mid", "-m", help="mid of the author", required=True, type=int)
@click.option(
    "--save-dir",
    "-s",
    help="Path to save the downloaded videos",
    type=Path,
    required=True,
)
@click.option("--sess-data", "-ss", help="`SESSDATA`", required=False)
@click.option("--max-workers", "-w", help="max workers", type=int, default=4)
def download_by_author(
    mid: int, save_dir: Path, sess_data: str | None = None, max_workers: int = 4
):
    with db.Session() as s:
        bvids = [
            row.bvid
            for row in s.execute(
                sa.select(db.BilibiliAuthorVideo.bvid).where(
                    db.BilibiliAuthorVideo.mid == mid
                )
            )
        ]

    print(
        f"[bold yellow]{len(bvids)} videos found to be downloaded for author {mid}..."
    )

    multiple_downloader = download.MultiThreadDownloader(
        bvid_list=bvids, save_dir=save_dir, sess_data=sess_data, max_workers=max_workers
    )

    multiple_downloader.download()


if __name__ == "__main__":
    cli()  # pragma: no cover



import logging
import os
import multiprocessing
from pathlib import Path
from environs import Env

BASE_DIR = Path(__file__).resolve().parent.parent.parent

WORKDIR_ENV = os.environ.get("WORKDIR", default=None)

WORKDIR = Path.cwd() if WORKDIR_ENV is None else Path(WORKDIR_ENV)

DOT_ENV = WORKDIR / ".env"

env = Env()

if DOT_ENV.exists():
    env.read_env(str(WORKDIR / ".env"))
else:
    env.read_env()

DEBUG = env.bool("DEBUG", False)

LOG_LEVEL = env.int("LOG_LEVEL", logging.DEBUG if DEBUG else logging.INFO)

LOG_DIR = WORKDIR / "logs"

DB_DIR = WORKDIR / ".db"

LOG_DIR.mkdir(exist_ok=True)

DB_DIR.mkdir(exist_ok=True)

CPU_COUNT = multiprocessing.cpu_count()


# NOTE: dm_img_str and dm_cover_img_str seem to be some kind of browser fingerprint, and they seem to be static values
# NOTE: Reference: https://github.com/SocialSisterYi/bilibili-API-collect/issues/868
BILIBILI_PARAM_DM_IMG_STR = env.str(
    "BILIBILI_PARAM_DM_IMG_STR", "V2ViR0wgMS4wIChPcGVuR0wgRVMgMi4wIENocm9taXVtKQ"
)
BILIBILI_PARAM_DM_COVER_IMG_STR = env.str(
    "BILIBILI_PARAM_DM_COVER_IMG_STR",
    "QU5HTEUgKEludGVsIEluYy4sIEludGVsKFIpIElyaXMoVE0pIFBsdXMgR3JhcGhpY3MgNjU1LCBPcGVuR0wgNC4xKUdvb2dsZSBJbmMuIChJbnRlbCBJbmMuKQ",
)



import json
import logging
import re
import subprocess
from bs4 import BeautifulSoup
from enum import Enum, auto
from operator import attrgetter
from pathlib import Path
from functools import cached_property
from typing import Callable, TypeAlias, Generator, BinaryIO, Any, Optional
from itertools import chain
from traceback import format_tb

from rich.progress import (
    Progress,
    DownloadColumn,
    TransferSpeedColumn,
    TaskID,
    TextColumn,
    SpinnerColumn,
    TimeElapsedColumn,
    BarColumn,
    TaskProgressColumn,
    TimeRemainingColumn,
    MofNCompleteColumn,
)
from concurrent import futures
from rich.console import Console
import requests

from spiders_for_all.bilibili import models
from spiders_for_all.utils.helper import user_agent_headers, rm_tree
from spiders_for_all.utils.logger import default_logger as logger
from spiders_for_all.conf import settings

Media: TypeAlias = models.PlayVideo | models.PlayAudio
Medias: TypeAlias = list[Media]
Videos: TypeAlias = list[models.PlayVideo]
Audios: TypeAlias = list[models.PlayAudio]
VideoPath: TypeAlias = Path
AudioPath: TypeAlias = Path
BvidList: TypeAlias = str | list[str] | Path | list[Path] | BinaryIO

API_PREFIX = "https://www.bilibili.com/video/"

# Regex to find playinfo in <script>window.__playinfo__=</script>
RGX_FIND_PLAYINFO = re.compile(r"<script>window\.__playinfo__=(.*?)</script>")
RGX_FIND_TITLE = re.compile(r"<title>(.*?)</title>")
RGX_CHECK_FILENAME = re.compile(r"[\\/:*?\"<>|]")

HIGHEST_QUALITY = 0

CHUNK_SIZE = 1024 * 1024

NOT_SET = object()


def get_ffmpeg_executable() -> str:
    return "ffmpeg"
    # FIXME: which may not work on windows
    # return subprocess.check_output(["which", "ffmpeg"]).decode().strip()


class MediaType(Enum):
    VIDEO = "video"
    AUDIO = "audio"


class DownloadState(Enum):
    NOT_STARTED = auto()
    STARTED = auto()
    FINISHED = auto()
    FAILED = auto()


class BaseTask:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._finished = False

    @property
    def finished(self) -> bool:
        return self._finished

    def start(self) -> Any:
        raise NotImplementedError()


class LinerTask(BaseTask):
    def __init__(
        self,
        fn: Callable,
        name: str | None = None,
        args: tuple | None = None,
        kwargs: dict | None = None,
        delay_args: Optional[Callable] = None,
        delay_kwargs: Optional[Callable] = None,
    ):
        """
        Liner task
        Args:
            fn: Function to call
            name: Name of the task
            args: Args to pass to the function
            kwargs: Kwargs to pass to the function
            delay_args: A callable object to return the args to pass to the function when calling
            delay_kwargs: A callable object to return the kwargs to pass to the function when calling

        About delay_args and delay_kwargs:

            Example:

                class A:

                    def __init__(self):
                        self.val = 1

                def fn_1(inst: A):
                    inst.val = 2

                def fn_2(val):
                    pass

                inst = A()

                task_1 = LinerTask(fn_1, args=(inst,))
                task_2 = LinerTask(fn_2, delay_args=lambda: (inst.val,))
                task_3 = LinerTask(fn_2, args=(inst.val,))

                task_1.start() # inst.val set to 2
                task_2.start() # will call fn_2 with new value 2
                task_3.start() # will call fn_2 with old value 1

        """
        self.fn = fn
        self.args = tuple() if args is None else args
        self.kwargs = {} if kwargs is None else kwargs
        self.task_name = name or str(fn)
        self.delay_args = delay_args
        self.delay_kwargs = delay_kwargs

        self.delay = any((self.delay_args, self.delay_kwargs))
        super().__init__()

    def start(self):
        if self.delay:
            if callable(self.delay_args):
                self.args = self.delay_args()
            if callable(self.delay_kwargs):
                self.kwargs = self.delay_kwargs()
        ret = self.fn(*self.args, **self.kwargs)
        self._finished = True
        return ret


class DownloadMediaTask(BaseTask):
    referer = "https://www.bilibili.com"

    def __init__(
        self,
        media: Media,
        save_dir: Path | str,
        *args,
        chunk_size: int = CHUNK_SIZE,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.media = media
        self.save_dir = Path(save_dir) if not isinstance(save_dir, Path) else save_dir
        self.chunk_size = chunk_size
        self.save_dir.unlink(missing_ok=True)
        self._total_size: int | None = None

    @property
    def total_size(self) -> int:
        if self._total_size is None:
            raise ValueError("Total size not fetched yet")
        return self._total_size

    @cached_property
    def headers(self) -> dict[str, str]:
        return {
            "Referer": self.referer,
            **user_agent_headers(),
        }

    def start(self) -> Generator[bytes | int, None, None]:
        with open(self.save_dir, "wb") as f:
            with requests.get(
                self.media.base_url, headers=self.headers, stream=True
            ) as resp:
                resp.raise_for_status()
                if self._total_size is None:
                    self._total_size = int(resp.headers["Content-Length"])
                    yield self._total_size
                for chunk in resp.iter_content(chunk_size=self.chunk_size):
                    yield chunk
                    f.write(chunk)

        self._finished = True

    def __str__(self):
        return str(self.media)


class DownloadVideoTask(DownloadMediaTask):
    def __init__(
        self,
        media: models.PlayVideo,
        save_dir: Path | str,
        quality_map: dict[int, str],
        chunk_size: int = CHUNK_SIZE,
    ):
        super().__init__(media, save_dir, chunk_size=chunk_size)
        self.quality_map = quality_map

    def __str__(self):
        return f"Download video <{self.quality_map[self.media.quality]}>"  # type: ignore


class DownloadAudioTask(DownloadMediaTask):
    def __str__(self):
        return "Download audio"


class Downloader:
    origin: str = "https://www.bilibili.com"
    api: str = "https://www.bilibili.com/video/{bvid}/"
    default_video_suffix: str = ".mp4"

    def __init__(
        self,
        bvid: str,
        save_dir: str | Path,
        filename: str | None = None,
        remove_temp_dir: bool = True,
        sess_data: str | None = None,
        quality: int = HIGHEST_QUALITY,
        codecs: str | None = None,
        ffmpeg_params: list[str] | None = None,
        process_func: Callable[[VideoPath, AudioPath], None] | None = None,
        from_cli: bool = True,
        exit_on_error: bool = False,
        ffmpeg: str | None = None,
        disable_terminal_log: bool = False,
    ):
        self.bvid = bvid
        self.save_dir = save_dir if isinstance(save_dir, Path) else Path(save_dir)
        self.temp_dir = self.save_dir / ".temp"

        self.audio_temp_path = self.temp_dir / "audio-temp.wav"
        self.remove_temp_dir = remove_temp_dir
        self.sess_data = sess_data
        self.quality = quality
        self.codecs = codecs
        self.ffmpeg_params = ffmpeg_params
        self.from_cli = from_cli

        self.process_func = process_func

        self.api = self.__class__.api.format(bvid=bvid)

        self.save_dir.mkdir(exist_ok=True, parents=True)
        self.temp_dir.mkdir(exist_ok=True, parents=True)
        self.log_file = self.get_logfile()

        self.audios_path: list[Path] = []
        self.temp_audio: Path | None = None
        self.video_path: Path | None = None

        self.tasks: list[LinerTask] = []
        self.downloading_tasks: list[DownloadMediaTask] = []

        self.exit_on_error = exit_on_error
        self.ffmpeg = ffmpeg or get_ffmpeg_executable()
        self.disable_terminal_log = disable_terminal_log
        self.console: Console | None = None

        self._filename = filename

        self.exc_info: tuple[Any, Any, Any] = (NOT_SET, NOT_SET, NOT_SET)
        self.state: DownloadState = DownloadState.NOT_STARTED

    def __enter__(self):
        self.state = DownloadState.STARTED

        self.console = self.get_console()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.state = DownloadState.FAILED
            self.exc_info = (exc_type, exc_val, exc_tb)
            self.log(
                f"[{self.downloader_name}] Error occurred: {exc_val}",
                level=logging.ERROR,
                traceback=exc_tb,
            )

        else:
            self.state = DownloadState.FINISHED
            self.log(
                f"[{self.downloader_name}] Download finished successfully",
                level=logging.INFO,
            )

        if self.console.record:  # type: ignore
            self.console.save_text(str(self.log_file))  # type: ignore

        return exc_type is None

    def get_console(self) -> Console:
        if self.disable_terminal_log:
            return Console(file=open(self.log_file, "w"))
        return Console(record=True)

    def get_logfile(self) -> Path:
        log_file = self.save_dir / f"{self.bvid}.log"
        log_file.touch(exist_ok=True)
        return log_file

    @cached_property
    def play_info(self) -> models.PlayInfoData:
        return self.get_play_info()

    @cached_property
    def filepath(self) -> Path:
        _filename = self._filename or self.get_title()

        # Replace invalid characters with _
        _filename = RGX_CHECK_FILENAME.sub("_", _filename)

        _filepath = self.save_dir / _filename

        if _filepath.suffix not in (".mp4", ".mkv", ".avi"):
            _filepath = _filepath.with_suffix(self.default_video_suffix)

        return _filepath

    @cached_property
    def html_content(self) -> str:
        headers = {**user_agent_headers()}
        cookies = {"SESSDATA": self.sess_data} if self.sess_data is not None else {}
        self.log(
            f"[{self.downloader_name}] Requesting {self.api}{f' with cookies {cookies}' if cookies else ''}"
        )
        response = requests.get(self.api, headers=headers, cookies=cookies)
        response.raise_for_status()
        html_content = response.text
        return html_content

    @cached_property
    def videos(self) -> Videos:
        """Get all videos sorted by quality in descending order"""
        return sorted(
            self.play_info.dash.video, key=attrgetter("quality"), reverse=True
        )

    @cached_property
    def video_to_download(self) -> models.PlayVideo:
        """Get the video to download according to quality and codecs"""
        return self.choose_codecs(self.filter_quality(self.videos))

    @cached_property
    def downloader_name(self) -> str:
        return self.__str__()

    def log(self, msg: str, level: int = logging.DEBUG, traceback: Any = None):
        if self.console is None:
            logger.log(level, msg, exc_info=traceback is not None)
        else:
            if level >= settings.LOG_LEVEL:
                self.console.log(f"[{logging.getLevelName(level)}] {msg}")
                if traceback:
                    self.console.log(f"Traceback: \n{''.join(format_tb(traceback))}")

    def get_play_info(self) -> models.PlayInfoData:
        playinfo = RGX_FIND_PLAYINFO.search(self.html_content)
        if playinfo is None:
            raise ValueError(f"Playinfo not found from {self.html_content}")
        playinfo = playinfo.group(1)  # type: ignore
        return models.PlayInfoResponse(**json.loads(playinfo)).data  # type: ignore

    def get_title(self) -> str:
        soup = BeautifulSoup(self.html_content, "html.parser")
        title_tag = soup.find("title")
        if title_tag is None:
            raise ValueError(f"Title not found from {self.html_content}")
        title = title_tag.text
        self.log(f"[{self.downloader_name}] Title: {title}", level=logging.INFO)
        return title

    def clean(self):
        if self.remove_temp_dir:
            # move file to the parent directory and remove save dir
            rm_tree(self.temp_dir)
            self.log(
                f"[{self.downloader_name}] Remove temporary directory",
                level=logging.INFO,
            )

        self.log(
            f"[{self.downloader_name}] Finished. File save to: {self.filepath}",
            level=logging.INFO,
        )

    def process(self, video_path: VideoPath, audio_path: AudioPath):
        """Process video and audio with ffmpeg

        Args:
            video_path (Path): video path
            audio_path (Path): audio path
        """
        # process with ffmpeg
        self.log(
            f"[{self.downloader_name}] Merging audio and video files to {self.filepath}",
            level=logging.INFO,
        )

        cmd = [
            self.ffmpeg,
            "-i",
            str(video_path),
            "-i",
            str(audio_path),
            "-c:v",
            "copy",
            "-c:a",
            "aac",
            str(self.filepath),
            "-y",
        ]

        if self.ffmpeg_params:
            cmd.extend(self.ffmpeg_params)

        self.log(
            f"[{self.downloader_name}] Process {video_path} with FFMPEG: {' '.join(cmd)}",
            level=logging.DEBUG,
        )

        result = subprocess.run(
            cmd,
            capture_output=True,
        )

        if result.returncode:
            raise ValueError(
                f"{cmd} failed with code {result.returncode}."
                f"Detail: {result.stderr.decode().strip()}"
            )

    def filter_quality(self, videos: Videos) -> Videos:
        if self.quality is HIGHEST_QUALITY:
            videos = videos[:1]
        else:
            videos = list(filter(lambda video: video.quality == self.quality, videos))
        if not videos:
            raise ValueError(
                f"No video with quality {self.quality} found, available qualities: {self.play_info.quality_map.items()}"
            )
        return videos

    def choose_codecs(self, videos: Videos) -> models.PlayVideo:
        if self.codecs is None:
            return videos[0]
        for video in videos:
            if re.search(self.codecs, video.codecs):
                return video

        codecs = ", ".join([video.codecs for video in videos])
        raise ValueError(
            f"No video with codec {self.codecs} matched, available codecs: {codecs}"
        )

    def run_task_directly(self, task: DownloadMediaTask):
        iterator = task.start()
        total_size: int = next(iterator)  # type: ignore
        for chunk in iterator:
            self.log(
                f"[{self.downloader_name}] {task}: {len(chunk) / total_size * 100:.2f}%",  # type: ignore
                level=logging.INFO,
            )

    def add_video_task(self):
        process_video_task = DownloadVideoTask(
            self.video_to_download,
            self.temp_dir
            / f"video-{self.video_to_download.quality}-{self.video_to_download.codecs}.mp4",
            self.play_info.quality_map,
        )
        self.downloading_tasks.append(process_video_task)

        self.video_path = process_video_task.save_dir

    def add_audio_task(self):
        for audio in self.play_info.dash.audio:
            self.downloading_tasks.append(
                DownloadAudioTask(
                    audio,
                    self.temp_dir / f"audio-{audio.audio_id}.mp4",
                )
            )

    def prepare_tasks(self):
        self.tasks.append(LinerTask(self.add_video_task, name="Search video links"))
        self.tasks.append(LinerTask(self.add_audio_task, name="Search audio links"))
        if self.from_cli:
            self.tasks.append(
                LinerTask(self.download_with_progress, name="Download video and audio")
            )
        else:
            self.tasks.append(
                LinerTask(self.download_directly, name="Download video and audio")
            )
        self.tasks.append(LinerTask(self.concat_audio, name="Merge audio files"))

        if self.process_func is not None and callable(self.process_func):
            self.tasks.append(
                LinerTask(
                    self.process_func,
                    name="Merge video and audio files",
                    delay_args=lambda: (self.video_path, self.audio_temp_path),
                )
            )
        else:
            self.tasks.append(
                LinerTask(
                    self.process,
                    name="Merge video and audio files",
                    delay_args=lambda: (self.video_path, self.audio_temp_path),
                )
            )

        self.tasks.append(LinerTask(self.clean, name="Clean up"))

    def run_tasks(self):
        total_tasks = len(self.tasks)
        for task_idx, task in enumerate(self.tasks):
            self.log(
                f"[{self.downloader_name}] [{task_idx + 1}/{total_tasks}]: {task.task_name}",
                level=logging.INFO,
            )
            task.start()

    def iter_tasks(self) -> Generator[LinerTask, None, None]:
        for task in self.tasks:
            task.start()
            yield task

    def download(
        self, yield_tasks: bool = False
    ) -> Optional[Generator[LinerTask, None, None]]:
        self.prepare_tasks()

        return self.run_tasks() if not yield_tasks else self.iter_tasks()

    @staticmethod
    def _update_progress(task: DownloadMediaTask, task_id: TaskID, progress: Progress):
        progress.start_task(task_id)
        while not task.finished:
            iterator = task.start()
            total_size = next(iterator)
            for chunk in iterator:
                progress.update(task_id, total=total_size, advance=len(chunk))  # type: ignore

    def download_with_progress(self):
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            DownloadColumn(),
            TransferSpeedColumn(),
        ) as progress:
            tasks_map = {
                task: progress.add_task(str(task), start=False)
                for task in self.downloading_tasks
            }
            with futures.ThreadPoolExecutor() as executor:
                tasks = {
                    executor.submit(
                        self._update_progress, task, task_id, progress
                    ): task
                    for task, task_id in tasks_map.items()
                }

                self.handle_tasks(tasks)

    def download_directly(self):
        with futures.ThreadPoolExecutor() as executor:
            tasks = {
                executor.submit(self.run_task_directly, task): task
                for task in self.downloading_tasks
            }

            self.handle_tasks(tasks)

    def concat_audio(self):
        with open(self.audio_temp_path, "wb") as f:
            for audio_task in filter(
                lambda t: isinstance(t, DownloadAudioTask), self.downloading_tasks
            ):
                f.write(audio_task.save_dir.read_bytes())

        self.log(
            f"[{self.downloader_name}] Audio file saved to {self.audio_temp_path}",
            level=logging.INFO,
        )

    def handle_tasks(self, tasks: dict[futures.Future, DownloadMediaTask]):
        for f in futures.as_completed(tasks):
            task = tasks[f]
            try:
                f.result()
            except Exception as e:
                self.log(
                    f"[{self.downloader_name}] Error occurred when downloading {task}: {e}",
                    level=logging.ERROR,
                )

                if self.exit_on_error:
                    exit(1)

    def __str__(self):
        return f"<Downloader {self.bvid}>"


class MultiThreadDownloader:
    rgx_split_bvid_list = re.compile(r"[\s,\t\n]+")

    def __init__(
        self,
        bvid_list: BvidList,
        save_dir: str | Path,
        remove_temp_dir: bool = True,
        sess_data: str | None = None,
        quality: int = HIGHEST_QUALITY,
        codecs: str | None = None,
        ffmpeg_params: list[str] | None = None,
        process_func: Callable[[VideoPath, AudioPath], None] | None = None,
        max_workers: int = settings.CPU_COUNT,
        from_cli: bool = True,
        exit_on_error: bool = False,
        ffmpeg: str | None = None,
    ):
        self.bvid_list = list(set(sorted(self.read_bvid_list(bvid_list))))
        self.save_dir = Path(save_dir) if not isinstance(save_dir, Path) else save_dir
        self.ffmpeg = ffmpeg or get_ffmpeg_executable()
        self.downloaders = [
            Downloader(
                bvid,
                self.save_dir / bvid,
                remove_temp_dir=remove_temp_dir,
                sess_data=sess_data,
                quality=quality,
                codecs=codecs,
                ffmpeg_params=ffmpeg_params,
                process_func=process_func,
                from_cli=False,
                ffmpeg=self.ffmpeg,
                disable_terminal_log=True,
                exit_on_error=True,
            )
            for bvid in self.bvid_list
        ]

        self.max_workers = max_workers
        self.from_cli = from_cli
        self.exit_on_error = exit_on_error
        self.log_dir = self.save_dir / "logs"
        self.log_dir.mkdir(exist_ok=True)

        self.success_count = 0
        self.failed_count = 0

    def read_bvid_list(self, bvid_list: BvidList) -> list[str]:
        match bvid_list:
            case str():
                return self.rgx_split_bvid_list.split(bvid_list.strip())
            case Path():
                return self.read_bvid_list(bvid_list.read_text())
            case BinaryIO():
                return self.read_bvid_list(bvid_list.read().decode())
            case list():
                return list(
                    chain.from_iterable(
                        filter(
                            lambda _bvid: _bvid,
                            map(
                                self.read_bvid_list,
                                bvid_list,
                            ),
                        )
                    )
                )
            case _:
                raise TypeError(
                    f"bvid_list should be str, Path, BinaryIO or list, got {type(bvid_list)}"
                )

    def download_directly(self):
        with futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            tasks = {
                executor.submit(downloader.download): downloader
                for downloader in self.downloaders
            }

            self.handle_tasks(tasks)

    def _worker_update_progress(
        self,
        downloader: Downloader,
        progress: Progress,
        overall_task_id: TaskID,
    ):
        downloader.prepare_tasks()
        total = len(downloader.tasks)
        task_id = progress.add_task(
            f"[green] {downloader.downloader_name}: ",
            event="Preparing",
            start=True,
            total=total,
        )

        try:
            with downloader:
                for idx, task in enumerate(downloader.iter_tasks()):
                    progress.update(
                        task_id,
                        completed=idx + 1,
                        total=total,
                        event=f"[green bold][{idx + 1}/{total}] {task.task_name}",
                    )
        except KeyboardInterrupt:
            exit(1)

        except:  # noqa: E722
            progress.update(task_id=task_id, event="[red bold] Failed!")
            self.failed_count += 1
        else:
            progress.update(overall_task_id, advance=1, event="")
            progress.update(task_id, event="[green bold] Finished!")
            self.success_count += 1
        finally:
            downloader.log_file = downloader.log_file.rename(
                self.log_dir / downloader.log_file.name
            )
            if downloader.state is DownloadState.FAILED:
                logger.error(
                    f"[{downloader.downloader_name}] Failed, please check the log file: {downloader.log_file} for detail",
                )
            else:
                downloader.filepath = downloader.filepath.rename(
                    self.save_dir / downloader.filepath.name
                )
                logger.info(
                    f"[{downloader.downloader_name}] Success, file saved to: {downloader.filepath}"
                )
            rm_tree(downloader.save_dir)
            progress.remove_task(task_id)
            progress.update(
                overall_task_id,
                event=f"[bold green]{self.success_count} succeeded {self.failed_count} failed",
            )

    def download_with_progress(self):
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            MofNCompleteColumn(),
            TaskProgressColumn(),
            TimeElapsedColumn(),
            TextColumn("{task.fields[event]}"),
            refresh_per_second=2,
        ) as progress:
            overall_task = progress.add_task(
                "[green] All jobs: ", total=len(self.downloaders), event=""
            )

            with futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                progress.start_task(overall_task)
                tasks = {
                    executor.submit(
                        self._worker_update_progress,
                        downloader,
                        progress,
                        overall_task,
                    ): downloader
                    for downloader in self.downloaders
                }

                self.handle_tasks(tasks)

        logger.info(f"All jobs finished, files saved to: {self.save_dir}")

    def download(self):
        if self.from_cli:
            self.download_with_progress()
        else:
            self.download_directly()

    def handle_tasks(
        self,
        tasks: dict[futures.Future, Downloader],
    ):
        for f in futures.as_completed(tasks):
            downloader = tasks[f]
            try:
                f.result()
            except Exception as e:
                logger.error(
                    f"Error occurred for {downloader}: {e}",
                    exc_info=True,
                )

                if self.exit_on_error:
                    exit(1)



import json
from typing import Type, TypeAlias
from functools import cached_property
from enum import Enum, auto

from rich import print
from rich import table as rich_table
from sqlalchemy import Row

from spiders_for_all.bilibili import db

ViewCount: TypeAlias = int
SortData: TypeAlias = tuple[int, tuple[ViewCount, Row]]  # row_id, (view_count, row)
VideoOrPlay: TypeAlias = db.BaseBilibiliVideos | db.BaseBilibiliPlay
VideoOrPlayClass: TypeAlias = Type[VideoOrPlay]
TableModelClass: TypeAlias = Type[db.Base]

N: int = 10


class ShowType(Enum):
    COMMON = auto()
    VIDEO_OR_PLAY = auto()


class Analysis:
    def __init__(self, model: TableModelClass, n: int = N) -> None:
        """_summary_

        Args:
            model (Model): model to analysis
            n (int, optional): N rows to analysis. Defaults to 10.

        Raises:
            TypeError: _description_
        """

        self.model = model
        self.n = n
        if not issubclass(model, (db.BaseBilibiliVideos, db.BaseBilibiliPlay)):
            self.table = self.get_table(self.get_model_columns(model))
            self.show_type = ShowType.COMMON
        else:
            self.table = rich_table.Table(
                "rank",
                "title",
                "view",
                "link",
                title=f"Top {n} videos",
                show_header=True,
                header_style="bold magenta",
                show_lines=True,
            )
            self.show_type = ShowType.VIDEO_OR_PLAY

    @staticmethod
    def get_model_columns(model: Type[db.Base]) -> list[str]:
        return model.__table__.columns.keys()

    def get_table(self, columns: list[str]) -> rich_table.Table:
        return rich_table.Table(
            *columns,
            title=f"Top {self.n} videos",
            show_header=True,
            header_style="bold magenta",
            show_lines=True,
        )

    def _key(self, videos_tuple: SortData):
        return videos_tuple[1][0]

    @cached_property
    def url_field(self) -> str:
        if issubclass(self.model, db.BaseBilibiliVideos):
            return "short_link_v2"
        return "url"

    def add_videos_or_plays_row(self, model: VideoOrPlayClass):
        with db.Session() as s:
            videos_data: dict[int, tuple[ViewCount, Row]] = {
                item.id: (json.loads(item.stat)["view"], item)
                for item in s.query(
                    model.id,
                    model.stat,
                    model.title,
                    getattr(self.model, self.url_field),
                )  # type: ignore
            }
            videos_data = dict(sorted(videos_data.items(), key=self._key))
            count = 0
            for i, (_, item_tuple) in enumerate(videos_data.items()):
                view_count, row = item_tuple
                if count >= self.n:
                    break

                self.table.add_row(
                    str(i + 1), row.title, str(view_count), getattr(row, self.url_field)
                )

                count += 1

    def add_common_row(self, model: TableModelClass):
        with db.Session() as s:
            for i, row in enumerate(s.query(model)):
                if i >= self.n:
                    break
                self.table.add_row(*[str(value) for value in row.tuple()])  # type: ignore

    def show(self):
        if self.show_type == ShowType.VIDEO_OR_PLAY:
            self.add_videos_or_plays_row(self.model)  # type: ignore
        else:
            self.add_common_row(self.model)

        print(self.table)



from datetime import datetime
import json
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.hybrid import hybrid_property
from spiders_for_all.bilibili import models
from spiders_for_all.conf import settings

db_file = settings.DB_DIR / "bilibili.db"

engine = sa.engine.create_engine(f"sqlite:///{str(db_file)}", echo=settings.DEBUG)

Session = orm.sessionmaker(bind=engine)

_INIT_DB = not db_file.exists()


class Base(orm.DeclarativeBase):
    pass


class BaseTable(Base):
    __abstract__ = True

    id: orm.Mapped[int] = orm.mapped_column(
        primary_key=True, comment="auto increment id"
    )

    create_at: orm.Mapped[datetime] = orm.mapped_column(
        default=datetime.now, comment="create time of a video"
    )

    update_at: orm.Mapped[datetime] = orm.mapped_column(
        default=datetime.now, comment="update time of a video", onupdate=datetime.now
    )

    def tuple(self):
        return tuple(
            [
                getattr(self, column.key)
                for column in sa.inspect(self).mapper.column_attrs
            ]
        )


class BaseBilibiliVideos(BaseTable):
    __abstract__ = True

    title: orm.Mapped[str] = orm.mapped_column(comment="title of a video")
    tname: orm.Mapped[str] = orm.mapped_column(comment="tag name of a video")

    aid: orm.Mapped[int] = orm.mapped_column(unique=True, comment="avid of a video")
    bvid: orm.Mapped[str] = orm.mapped_column(unique=True, comment="bvid of a video")
    cid: orm.Mapped[int] = orm.mapped_column(comment="cid")
    desc: orm.Mapped[str] = orm.mapped_column(comment="description of a video")
    owner: orm.Mapped[str] = orm.mapped_column(comment="owner info of a video")
    pubdate: orm.Mapped[int] = orm.mapped_column(comment="pubdate of a video")
    short_link_v2: orm.Mapped[str] = orm.mapped_column(
        comment="short_link_v2 of a video"
    )
    stat: orm.Mapped[str] = orm.mapped_column(comment="stat of a video")
    tid: orm.Mapped[int] = orm.mapped_column(comment="tid of a video")

    @hybrid_property
    def owner_info(self) -> models.VideoOwner:
        return models.VideoOwner(**json.loads(self.owner))

    @hybrid_property
    def stat_info(self) -> models.VideoStat:
        return models.VideoStat(**json.loads(self.stat))

    @hybrid_property
    def url(self) -> str:
        return self.short_link_v2


class BaseBilibiliPlay(Base):
    __abstract__ = True

    id: orm.Mapped[int] = orm.mapped_column(
        comment="auto increment id", primary_key=True
    )
    rank: orm.Mapped[int] = orm.mapped_column(comment="rank of a video")
    rating: orm.Mapped[str] = orm.mapped_column(comment="rating of a video")
    stat: orm.Mapped[str] = orm.mapped_column(comment="stat of a video")
    title: orm.Mapped[str] = orm.mapped_column(comment="title of a video")
    url: orm.Mapped[str] = orm.mapped_column(comment="url of a video")

    create_at: orm.Mapped[datetime] = orm.mapped_column(
        default=datetime.now, comment="create time of a video"
    )

    update_at: orm.Mapped[datetime] = orm.mapped_column(
        default=datetime.now, comment="update time of a video", onupdate=datetime.now
    )


class BilibiliPopularVideos(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_popular_videos"
    __doc__ = "bilibili popular videos information"


class BilibiliWeeklyVideos(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_weekly_videos"
    __doc__ = "bilibili weekly videos information"

    aid: orm.Mapped[int] = orm.mapped_column(comment="avid of a video")
    bvid: orm.Mapped[str] = orm.mapped_column(comment="bvid of a video")
    week: orm.Mapped[int] = orm.mapped_column(comment="Which week of a video")


class BilibiliPreciousVideos(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_precious_videos"
    __doc__ = "bilibili precious videos information"

    achievement: orm.Mapped[str]
    bvid: orm.Mapped[str] = orm.mapped_column(comment="bvid of a video")
    aid: orm.Mapped[int] = orm.mapped_column(comment="avid of a video")


class BilibiliRankDrama(BaseBilibiliPlay):
    __tablename__ = "t_bilibili_rank_drama"
    __doc__ = "bilibili rank drama information"


class BilibiliRankAll(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_all"
    __doc__ = "bilibili rank all information"


class BilibiliRankCnCartoon(BaseBilibiliPlay):
    __tablename__ = "t_bilibili_rank_cn_cartoon"
    __doc__ = "bilibili rank cn cartoon information"


class BilibiliRankCnRelated(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_cn_related"
    __doc__ = "bilibili rank cn related information"


class BilibiliRankDocumentary(BaseBilibiliPlay):
    __tablename__ = "t_bilibili_rank_documentary"
    __doc__ = "bilibili rank documentary information"


class BilibiliRankCartoon(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_cartoon"
    __doc__ = "bilibili rank cartoon information"


class BilibiliRankMusic(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_music"
    __doc__ = "bilibili rank music information"


class BilibiliRankDance(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_dance"
    __doc__ = "bilibili rank dance information"


class BilibiliRankGame(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_game"
    __doc__ = "bilibili rank game information"


class BilibiliRankTech(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_tech"
    __doc__ = "bilibili rank tech information"


class BilibiliRankKnowledge(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_knowledge"
    __doc__ = "bilibili rank knowledge information"


class BilibiliRankSport(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_sport"
    __doc__ = "bilibili rank sport information"


class BilibiliRankCar(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_car"
    __doc__ = "bilibili rank car information"


class BilibiliRankLife(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_life"
    __doc__ = "bilibili rank life information"


class BilibiliRankFood(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_food"
    __doc__ = "bilibili rank food information"


class BilibiliRankAnimal(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_animal"
    __doc__ = "bilibili rank animal information"


class BilibiliRankAuto(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_auto"
    __doc__ = "bilibili rank auto information"


class BilibiliRankFashion(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_fashion"
    __doc__ = "bilibili rank fashion information"


class BilibiliRankEnt(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_ent"
    __doc__ = "bilibili rank ent information"


class BilibiliRankFilm(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_film"
    __doc__ = "bilibili rank film information"


class BilibiliRankMovie(BaseBilibiliPlay):
    __tablename__ = "t_bilibili_rank_movie"
    __doc__ = "bilibili rank movie information"


class BilibiliRankTv(BaseBilibiliPlay):
    __tablename__ = "t_bilibili_rank_tv"
    __doc__ = "bilibili rank tv information"


class BilibiliRankVariety(BaseBilibiliPlay):
    __tablename__ = "t_bilibili_rank_variety"
    __doc__ = "bilibili rank variety information"


class BilibiliRankOrigin(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_origin"
    __doc__ = "bilibili rank origin information"


class BilibiliRankNew(BaseBilibiliVideos):
    __tablename__ = "t_bilibili_rank_new"
    __doc__ = "bilibili rank new information"


class BilibiliAuthorVideo(BaseTable):
    __tablename__ = "t_bilibili_author_video"
    __doc__ = "bilibili author video information"

    title: orm.Mapped[str] = orm.mapped_column(comment="title of a video")
    aid: orm.Mapped[int] = orm.mapped_column(comment="avid of a video", unique=True)
    bvid: orm.Mapped[str] = orm.mapped_column(comment="bvid of a video", unique=True)
    mid: orm.Mapped[int] = orm.mapped_column(comment="mid of a video")
    comment: orm.Mapped[int] = orm.mapped_column(comment="comment of a video")
    description: orm.Mapped[str] = orm.mapped_column(comment="description of a video")
    is_pay: orm.Mapped[int] = orm.mapped_column(comment="is_pay of a video")
    length: orm.Mapped[int] = orm.mapped_column(comment="length of a video")


def init_db():
    """init database"""
    Base.metadata.drop_all(engine)  # pragma: no cover
    Base.metadata.create_all(engine)  # pragma: no cover


if _INIT_DB:
    init_db()



from __future__ import annotations

import typing
import time
import hashlib
import random
from typing import TypeAlias
from functools import cached_property
from itertools import batched

import requests
import sqlalchemy as sa
from pydantic import BaseModel
from sqlalchemy import orm

from spiders_for_all.bilibili import models
from spiders_for_all.bilibili import db
from spiders_for_all.core.base import SPIDERS, Spider, get_all_spiders
from spiders_for_all.utils import helper
from spiders_for_all.utils.logger import get_logger
from spiders_for_all.conf import settings

Session = db.Session

_BilibiliResponse: TypeAlias = (
    models.BilibiliPlayResponse | models.BilibiliVideoResponse
)
_BilibiliResponseTyped: TypeAlias = typing.Type[_BilibiliResponse]

logger = get_logger("bilibili")


def calculate_end_page(total: int, page_size: int, start_page_number: int) -> int:
    """
    Calculate the end page number

    :param total: Total number of items to be crawled
    :param page_size: Number of items per page
    :param start_page_number: Start page number
    :return: End page number
    """
    return (
        start_page_number + (total // page_size) + (1 if total % page_size else 0) - 1
    )


class BaseBilibiliSpider(Spider):
    response_model: _BilibiliResponseTyped
    logger = logger
    insert_batch = 100

    def before(self):
        """
        Called before the spider starts
        """
        self.info(f"[{'Start':^10}]: {self.name}[{self.alias}]")

    def after(self):
        """
        Called after the spider finishes
        """
        self.info(f"[{'Finished':^10}]: {self.name}[{self.alias}]")

    def get_items(self) -> typing.Iterable[BaseModel]:
        """
        Get items from response
        :return: An iterable object of `pydantic.BaseModel`
        """
        yield from self.get_items_from_response(self.send_request())

    def process_item(self, item: BaseModel, **kwargs) -> orm.DeclarativeBase:
        """
        Create sqlalchemy model from pydantic model and kwargs
        :param item: item validated by pydantic model
        :param kwargs: extra kwargs to save to database
        :return:
        """
        return self.database_model(
            **{
                **item.model_dump(),
                **kwargs,
            }
        )

    def send_request(
        self, method: str = "GET"
    ) -> models.BilibiliVideoResponse | models.BilibiliPlayResponse:
        """
        Send request to api
        :param method: http method
        :return: A `pydantic.BaseModel` object
        """
        if not issubclass(self.response_model, models.BilibiliResponse):
            raise TypeError(
                f"{self.response_model} is not a valid response model, "
                f"should be a subclass of {models.BilibiliResponse}"
            )
        req_kwargs = self.get_request_args()
        self.debug(f"==> {method.upper()} {self.api}: {req_kwargs}")
        resp = requests.request(method, self.api, **req_kwargs)
        self.debug(f"<== {method.upper()} {resp.request.url}: {resp.status_code}")
        resp.raise_for_status()
        json_data = resp.json()
        self.debug(f"<== Response json: {json_data}")
        return self.response_model(**json_data)

    def get_request_args(self) -> dict:
        """
        Request kwargs to set
        """
        return {"headers": helper.user_agent_headers()}

    def get_items_from_response(
        self, response: _BilibiliResponse
    ) -> typing.Iterable[BaseModel]:
        """
        Get list data items from response model
        :param response:
        """
        response.raise_for_status()
        return response.data.list_data

    def save_to_db(self, items: typing.Iterable[BaseModel]):
        """
        Save items to database
        :param items: An iterable object of `pydantic.BaseModel`
        :return:
        """
        with Session() as s:
            for batch in batched(items, self.insert_batch):
                self.recreate(batch, session=s)
                self.info(f"Inserted {len(batch)} items")
            s.commit()

    def recreate(self, items: typing.Iterable[BaseModel], session: orm.Session):
        """
        Upsert/Recreate items to database
        :param items: items to be upserted
        :param session: sqlalchemy session object
        :return:
        """

        # TODO: Make truncate and upsert as options
        session.execute(sa.delete(self.database_model))
        session.add_all(map(self.process_item, items))

    def run(self):
        """
        Run the spider
        """
        self.before()
        self.save_to_db(self.get_items())
        self.after()


class BasePageSpider(BaseBilibiliSpider):
    default_page_size: int = 20
    default_page_number: int = 1
    default_total: int = 100

    def __init__(
        self,
        total: int = default_total,
        page_size: int = default_page_size,
        page_number: int = default_page_number,
        sleep_before_next_request: float | int | tuple[int, int] | None = None,
    ):
        """
        :param total: Total number of items to be crawled
        :param page_size: Number of items per page
        :param page_number: Start page number
        """
        super().__init__()
        self.total = int(total)
        self.page_size = int(page_size)
        self.page_number = int(page_number)
        self.end_page_number = calculate_end_page(
            self.total, self.page_size, self.page_number
        )
        self.sleep_before_next_request = sleep_before_next_request

    def get_items(self) -> typing.Iterable[BaseModel]:
        """
        Get items by page
        :return:
        """
        count = 0
        while self.page_number <= self.end_page_number:
            # TODO: change page size dynamically instead of calculating `count`
            response = self.send_request()
            items = self.get_items_from_response(response)
            if not items:
                break
            for item in items:
                yield item
                count += 1
                if count >= self.total:
                    return
            if self.sleep_before_next_request is not None:
                time.sleep(random.randrange(2, 5))
            self.page_number += 1

    def sleep(self):
        match self.sleep_before_next_request:
            case int():
                time.sleep(float(self.sleep_before_next_request))
            case float():
                time.sleep(self.sleep_before_next_request)
            case tuple():
                start, end = self.sleep_before_next_request
                time.sleep(random.randrange(start=start, stop=end))
            case _:
                raise TypeError(f"Invalid type: {type(self.sleep_before_next_request)}")


class BaseSearchSpider(BaseBilibiliSpider):
    def __init__(self, **search_params):
        super().__init__()
        self.search_params = search_params


class PopularSpider(BasePageSpider):
    api: str = "https://api.bilibili.com/x/web-interface/popular"
    name: str = "popular"
    alias = "综合热门"
    database_model = db.BilibiliPopularVideos
    item_model = models.PopularVideoItem
    response_model = models.PopularResponse

    def get_request_args(self) -> dict:
        return {
            **super().get_request_args(),
            "params": {
                "pn": self.page_number,
                "ps": self.page_size,
            },
        }

    def recreate(self, items: typing.Iterable[BaseModel], session: orm.Session):
        items = list(items)
        session.execute(
            sa.delete(self.database_model).where(
                self.database_model.bvid.in_([item.bvid for item in items]),  # type: ignore
            )
        )
        session.add_all([self.process_item(item) for item in items])


class WeeklySpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/popular/series/one"
    api_list = "https://api.bilibili.com/x/web-interface/popular/series/list"
    name = "weekly"
    alias = "每周必看"
    item_model = models.WeeklyVideoItem
    response_model = models.WeeklyResponse
    database_model = db.BilibiliWeeklyVideos

    search_key: str = "week"

    numbers_list: list[int] | None = None

    @cached_property
    def week(self) -> int:
        # Use the latest week number if not specified
        if self.search_key not in self.search_params:
            self.get_all_numbers()
            return max(self.numbers_list)  # type: ignore
        return int(self.search_params[self.search_key])

    def get_request_args(self) -> dict:
        return {**super().get_request_args(), "params": {"number": self.week}}

    def recreate(self, items: typing.Iterable[BaseModel], session: orm.Session):
        session.execute(
            sa.delete(self.database_model).where(
                self.database_model.week == self.week,  # type: ignore
            )
        )
        session.add_all([self.process_item(item, week=self.week) for item in items])

    @classmethod
    def get_all_numbers(cls) -> list[int]:
        if cls.numbers_list is None:
            resp = requests.get(
                cls.api_list,
                headers={
                    **helper.user_agent_headers(),
                },
            )
            resp.raise_for_status()
            cls.numbers_list = [item["number"] for item in resp.json()["data"]["list"]]
        return cls.numbers_list


class PreciousSpider(BasePageSpider):
    api = "https://api.bilibili.com/x/web-interface/popular/precious"
    name = "precious"
    alias = "入站必刷"
    default_page_size = 100

    item_model = models.PreciousVideoItem
    response_model = models.PreciousResponse
    database_model = db.BilibiliPreciousVideos

    def get_request_args(self) -> dict:
        return {
            **super().get_request_args(),
            "params": {"page_size": self.page_size, "page": self.page_number},
        }


class RankAllSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=0&type=all"
    name = "rank_all"
    alias = "全站"

    item_model = models.VideoItem
    database_model = db.BilibiliRankAll
    response_model = models.BilibiliVideoResponse


class RankDramaSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/pgc/web/rank/list?day=3&season_type=1"
    name = "rank_drama"
    alias = "番剧"

    item_model = models.PlayItem
    database_model = db.BilibiliRankDrama
    response_model = models.RankDramaResponse


class RankCnCartoonSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/pgc/season/rank/web/list?day=3&season_type=4"
    name = "rank_cn_cartoon"
    alias = "国产动画"

    item_model = models.PlayItem
    database_model = db.BilibiliRankCnCartoon
    response_model = models.BilibiliPlayResponse


class RankCnRelatedSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=168&type=all"
    name = "rank_cn_related"
    alias = "国创相关"

    item_model = models.VideoItem
    database_model = db.BilibiliRankCnRelated
    response_model = models.BilibiliVideoResponse


class RankDocumentarySpider(BaseSearchSpider):
    api = "https://api.bilibili.com/pgc/season/rank/web/list?day=3&season_type=3"
    name = "rank_documentary"
    alias = "纪录片"

    item_model = models.PlayItem
    database_model = db.BilibiliRankDocumentary
    response_model = models.BilibiliPlayResponse


class RankCartoonSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=1&type=all"
    name = "rank_cartoon"
    alias = "动画"

    item_model = models.VideoItem
    database_model = db.BilibiliRankCartoon
    response_model = models.BilibiliVideoResponse


class RankMusicSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=3&type=all"
    name = "rank_music"
    alias = "音乐"

    item_model = models.VideoItem
    database_model = db.BilibiliRankMusic
    response_model = models.BilibiliVideoResponse


class RankDanceSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=129&type=all"
    name = "rank_dance"
    alias = "舞蹈"

    item_model = models.VideoItem
    database_model = db.BilibiliRankDance
    response_model = models.BilibiliVideoResponse


class RankGameSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=4&type=all"
    name = "rank_game"
    alias = "游戏"

    item_model = models.VideoItem
    database_model = db.BilibiliRankGame
    response_model = models.BilibiliVideoResponse


class RankTechSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=188&type=all"
    name = "rank_tech"
    alias = "科技"

    item_model = models.VideoItem
    database_model = db.BilibiliRankTech
    response_model = models.BilibiliVideoResponse


class RankKnowledgeSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=36&type=all"
    name = "rank_knowledge"
    alias = "知识"

    item_model = models.VideoItem
    database_model = db.BilibiliRankKnowledge
    response_model = models.BilibiliVideoResponse


class RankSportSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=234&type=all"
    name = "rank_sport"
    alias = "运动"

    item_model = models.VideoItem
    database_model = db.BilibiliRankSport
    response_model = models.BilibiliVideoResponse


class RankCarSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=223&type=all"
    name = "rank_car"
    alias = "汽车"

    item_model = models.VideoItem
    database_model = db.BilibiliRankCar
    response_model = models.BilibiliVideoResponse


class RankLifeSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=160&type=all"
    name = "rank_life"
    alias = "生活"

    item_model = models.VideoItem
    database_model = db.BilibiliRankLife
    response_model = models.BilibiliVideoResponse


class RankFoodSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=211&type=all"
    name = "rank_food"
    alias = "美食"

    item_model = models.VideoItem
    database_model = db.BilibiliRankFood
    response_model = models.BilibiliVideoResponse


class RankAnimalSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=217&type=all"
    name = "rank_animal"
    alias = "动物圈"

    item_model = models.VideoItem
    database_model = db.BilibiliRankAnimal
    response_model = models.BilibiliVideoResponse


class RankAutoTuneSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=119&type=all"
    name = "rank_auto_tune"
    alias = "鬼畜"

    item_model = models.VideoItem
    database_model = db.BilibiliRankAuto
    response_model = models.BilibiliVideoResponse


class RankFashionSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=155&type=all"
    name = "rank_fashion"
    alias = "时尚"

    item_model = models.VideoItem
    database_model = db.BilibiliRankFashion
    response_model = models.BilibiliVideoResponse


class RankEntSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=5&type=all"
    name = "rank_ent"
    alias = "娱乐"

    item_model = models.VideoItem
    database_model = db.BilibiliRankEnt
    response_model = models.BilibiliVideoResponse


class RankFilmSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=181&type=all"
    name = "rank_film"
    alias = "影视"

    item_model = models.VideoItem
    database_model = db.BilibiliRankFilm
    response_model = models.BilibiliVideoResponse


class RankMovieSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/pgc/season/rank/web/list?day=3&season_type=2"
    name = "rank_movie"
    alias = "电影"

    item_model = models.PlayItem
    database_model = db.BilibiliRankMovie
    response_model = models.BilibiliPlayResponse


class RankTvSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/pgc/season/rank/web/list?day=3&season_type=5"
    name = "rank_tv"
    alias = "电视剧"

    item_model = models.PlayItem
    database_model = db.BilibiliRankTv
    response_model = models.BilibiliPlayResponse


class RankVarietySpider(BaseSearchSpider):
    api = "https://api.bilibili.com/pgc/season/rank/web/list?day=3&season_type=7"
    name = "rank_variety"
    alias = "综艺"

    item_model = models.PlayItem
    database_model = db.BilibiliRankVariety
    response_model = models.BilibiliPlayResponse


class RankOriginSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=0&type=origin"
    name = "rank_origin"
    alias = "原创"

    item_model = models.VideoItem
    database_model = db.BilibiliRankOrigin
    response_model = models.BilibiliVideoResponse


class RankNewSpider(BaseSearchSpider):
    api = "https://api.bilibili.com/x/web-interface/ranking/v2?rid=0&type=rookie"
    name = "rank_new"
    alias = "新人"

    item_model = models.VideoItem
    database_model = db.BilibiliRankNew
    response_model = models.BilibiliVideoResponse


class AuthorSpider(BasePageSpider):
    api = "https://api.bilibili.com/x/space/wbi/arc/search"
    api_get_nav_num = "https://api.bilibili.com/x/space/navnum"
    api_get_nav = "https://api.bilibili.com/x/web-interface/nav"
    name = "author"
    alias = "up主"

    response_model = models.AuthorVideoResponse
    database_model = db.BilibiliAuthorVideo
    item_model = models.AuthorVideoItem

    encrypt_string_fmt = "dm_cover_img_str={dm_cover_img_str}&dm_img_list=%5B%5D&dm_img_str={dm_img_str}&keyword=&mid={mid}&order=pubdate&order_avoided=true&platform=web&pn={pn}&ps={ps}&tid=&web_location="

    def __init__(
        self,
        mid: int,
        total: int | None = None,
        sess_data: str | None = None,
        page_size: int = 30,
        page_number: int = 1,
    ):
        self.mid = mid

        self.sess_data = sess_data

        self.wbi_info = self.get_wbi_info()

        self.key = self.get_mixin_key(self.wbi_info.img_key + self.wbi_info.sub_key)

        total = (
            self.get_total(self.mid)
            if total is None
            else min(total, self.get_total(self.mid))
        )

        super().__init__(
            total=total,
            page_size=page_size,
            page_number=page_number,
            sleep_before_next_request=(5, 10),
        )

    def get_wbi_info(self) -> models.WbiInfo:
        resp = requests.get(
            self.api_get_nav,
            headers=helper.user_agent_headers(),
        )

        resp.raise_for_status()

        wbi_img = resp.json().get("data", {}).get("wbi_img", None)

        if wbi_img is None:
            raise ValueError("wbi_img not found")

        return models.WbiInfo(**wbi_img)

    def get_mixin_key(self, e: str) -> str:
        indices = [
            46,
            47,
            18,
            2,
            53,
            8,
            23,
            32,
            15,
            50,
            10,
            31,
            58,
            3,
            45,
            35,
            27,
            43,
            5,
            49,
            33,
            9,
            42,
            19,
            29,
            28,
            14,
            39,
            12,
            38,
            41,
            13,
            37,
            48,
            7,
            16,
            24,
            55,
            40,
            61,
            26,
            17,
            0,
            1,
            60,
            51,
            30,
            4,
            22,
            25,
            54,
            21,
            56,
            59,
            6,
            63,
            57,
            62,
            11,
            36,
            20,
            34,
            44,
            52,
        ]
        result = []
        for r in indices:
            if r < len(e):
                result.append(e[r])
        return "".join(result)[:32]

    def get_total(self, mid: int) -> int:
        resp = requests.get(
            self.api_get_nav_num,
            params={"mid": mid},
            headers=helper.user_agent_headers(),
        )
        resp.raise_for_status()
        total = resp.json().get("data", {}).get("video", None)
        if total is None:
            raise ValueError(f"mid {mid} info not found")
        return total

    def get_items_from_response(
        self,
        # TODO: fix mypy error
        response: models.AuthorVideoResponse,  # type: ignore
    ) -> typing.Iterable[models.AuthorVideoItem]:
        response.raise_for_status()
        return response.data.list_data.items

    def get_wrid(self, params: str) -> str:
        md5 = hashlib.md5()

        md5.update(
            (params + self.key).encode(),
        )

        return md5.hexdigest()

    def get_request_args(self) -> dict:
        wts = round(time.time())

        params = self.encrypt_string_fmt.format(
            mid=self.mid,
            ps=self.page_size,
            pn=self.page_number,
            dm_img_str=settings.BILIBILI_PARAM_DM_IMG_STR,
            dm_cover_img_str=settings.BILIBILI_PARAM_DM_COVER_IMG_STR,
        )

        wrid = self.get_wrid(params + f"&wts={wts}")

        params = "&".join([params, f"w_rid={wrid}", f"wts={wts}"])

        kwargs = super().get_request_args()

        # NOTE: Not sure if this will work
        kwargs["headers"]["accept-language"] = "en,zh-CN;q=0.9,zh;q=0.8"

        if self.sess_data is not None:
            kwargs["cookies"] = {"SESSDATA": self.sess_data}

        return {
            **kwargs,
            "params": params,
        }

    def recreate(self, items: typing.Iterable[BaseModel], session: orm.Session):
        items = list(items)
        session.execute(
            sa.delete(self.database_model).where(
                self.database_model.bvid.in_([item.bvid for item in items]),  # type: ignore
            )
        )
        session.add_all([self.process_item(item) for item in items])


def run_spider(name: str, *args, **kwargs):
    try:
        spider = SPIDERS[name](*args, **kwargs)
    except KeyError as ke:
        print(f"spider not found: {ke.args[0]}")
        exit(1)
    spider.run()


if __name__ == "__main__":
    spiders = get_all_spiders()
    for spider_cls, spider_item in spiders.items():
        _spider = spider_cls()
        _spider.run()



from __future__ import annotations

import abc
import logging
import typing
from collections import namedtuple

from pydantic import BaseModel
from sqlalchemy import orm

from spiders_for_all.conf import settings
from spiders_for_all.utils.logger import default_logger

SPIDERS: dict[str, typing.Type[Spider]] = {}

SpiderItem = namedtuple("SpiderItem", ["name", "alias", "spider"])


def get_all_spiders() -> dict[typing.Type[Spider], SpiderItem]:
    return {
        spider: SpiderItem(name=spider.name, alias=spider.alias, spider=spider)
        for spider in set(SPIDERS.values())
    }


class Spider(abc.ABC):
    api: str
    name: str
    alias: str

    database_model: typing.Type[orm.DeclarativeBase]
    item_model: typing.Type[BaseModel]
    response_model: typing.Type[BaseModel]

    logger: logging.Logger = default_logger

    def __init__(self, *args, **kwargs):
        self.logger = kwargs.get("logger", self.__class__.logger)

        self.check_implementation()

    def check_implementation(self):
        attrs_required = [
            "api",
            "name",
            "alias",
            "database_model",
            "item_model",
            "response_model",
        ]

        for attr in attrs_required:
            if getattr(self, attr, None) is None:
                raise ValueError(f"Attribute {attr} is required")

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "name"):
            if cls.name in SPIDERS:
                raise ValueError(f"Duplicate spider name: {cls.name}")
            SPIDERS[cls.name] = cls
        if hasattr(cls, "alias"):
            if cls.alias in SPIDERS:
                raise ValueError(f"Duplicate spider alias: {cls.alias}")
            SPIDERS[cls.alias] = cls

    def log(self, msg: str, level: int = settings.LOG_LEVEL):
        self.logger.log(level, msg)

    def debug(self, msg: str):
        self.logger.debug(msg)

    def info(self, msg: str):
        self.logger.info(msg)

    def warning(self, msg: str):
        self.logger.warning(msg)

    def error(self, msg: str):
        self.logger.error(msg, exc_info=True)

    def critical(self, msg: str):
        self.logger.critical(msg)

    @abc.abstractmethod
    def run(self):
        pass

    @classmethod
    def string(cls) -> str:
        return f"<Spider {cls.name}({cls.alias}). Database model: {cls.database_model}>"
